import uuid
import msgpack
import threading
import traceback
from collections.abc import Callable
from typing import Any
from websocket import WebSocketApp, WebSocket
from dataclasses import dataclass
from .... import utils
from ....packets import Packet_CommandOutput
from ....utils import fmts


class MessageType:
    CMD_SET_SERVER_PKTS = "SetServerListenPackets"
    CMD_SET_CLIENT_PKTS = "SetClientListenPackets"
    CMD_SET_SERVER_BLOCK_PKTS = "SetBlockingServerPackets"
    CMD_SET_CLIENT_BLOCK_PKTS = "SetBlockingClientPackets"
    MSG_SERVER_PKT = "ServerMCPacket"
    MSG_CLIENT_PKT = "ClientMCPacket"
    MSG_SET_BOT_BASIC_INFO = "SetBotBasicInfo"
    MSG_UPDATE_UQ = "UpdateUQ"


@dataclass
class Message:
    type: str
    content: Any

    def dumps(self):
        return {"type": self.type, "content": self.content}


@dataclass
class PlayerUQ:
    name: str
    uuid: str
    xuid: str
    uniqueID: int
    abilities: dict


class Eulogist:
    """赞颂者启动器核心"""

    packet_listener: Callable[[int, dict], None] | None = None
    connected = False
    bot_data_ready_event = threading.Event()
    uq_data_ready_event = threading.Event()
    launch_event = threading.Event()
    exit_event = threading.Event()

    def __init__(self) -> None:
        self.command_cbs: dict[bytes, Callable[[dict], None]] = {}
        self.bot_name = ""
        self.bot_unique_id = 0
        self.bot_runtime_id = 0
        self.bot_uuid = ""
        self.uqs: dict[str, PlayerUQ] = {}
        self.client_packet_handler: Callable[[int, dict], None] = lambda x, y: None

    @staticmethod
    def make_conn(
        ipaddr: str,
        on_conn: Callable[[WebSocket], None],
        on_msg: Callable[[WebSocket, str], None],
        on_close: Callable[[WebSocket, Any, Any], None],
    ):
        return WebSocketApp(
            ipaddr, on_open=on_conn, on_message=on_msg, on_close=on_close
        )

    def start(self):
        self.conn = self.make_conn(
            "ws://127.0.0.1:10132", self.on_conn, self.on_msg, self.on_clos
        )
        self.conn.run_forever()

    def on_conn(self, ws: WebSocket):
        self.connected = True
        self.on_wait()

    @utils.thread_func("等待赞颂者连接初始化")
    def on_wait(self):
        # 需要一些时间
        fmts.print_inf("正在接收机器人基本信息...")
        self.bot_data_ready_event.wait()
        fmts.print_inf("正在接收在线玩家权限能力信息...")
        self.uq_data_ready_event.wait()
        self.launch_event.set()

    def on_msg(self, ws: WebSocket, msg_raw: str):
        msgdata = msgpack.unpackb(msg_raw)
        # print(msgdata)
        try:
            self.handler(Message(msgdata["type"], msgdata["content"]))
        except Exception:
            fmts.print_err("赞颂者发送的消息解析失败")
            fmts.print_err(traceback.format_exc())

    def on_clos(self, ws: WebSocket, _, _2):
        self.connected = False
        fmts.print_inf("赞颂者和 ToolDelta 断开连接")
        self.exit_event.set()

    def set_listen_server_packets(self, pkIDs: list[int]):
        self.send(Message(MessageType.CMD_SET_SERVER_PKTS, {"PacketsID": pkIDs}))

    def set_listen_client_packets(self, pkIDs: list[int]):
        self.send(Message(MessageType.CMD_SET_CLIENT_PKTS, {"PacketsID": pkIDs}))

    def set_blocking_server_packets(self, pkIDs: list[int]):
        self.send(Message(MessageType.CMD_SET_SERVER_BLOCK_PKTS, {"PacketsID": pkIDs}))

    def set_blocking_client_packets(self, pkIDs: list[int]):
        self.send(Message(MessageType.CMD_SET_CLIENT_BLOCK_PKTS, {"PacketsID": pkIDs}))

    def sendPacket(self, pkID: int, pk: dict):
        pk_bytes: Any = msgpack.packb(pk)
        self.send(
            Message(MessageType.MSG_SERVER_PKT, {"ID": pkID, "Content": pk_bytes})
        )

    def sendClientPacket(self, pkID: int, pk: dict):
        pk_bytes: Any = msgpack.packb(pk)
        self.send(
            Message(MessageType.MSG_CLIENT_PKT, {"ID": pkID, "Content": pk_bytes})
        )

    def sendcmd(self, cmd: str):
        u = uuid.uuid4()
        self.sendPacket(
            77,
            {
                "CommandLine": cmd,
                "CommandOrigin": {
                    "Origin": 0,
                    "UUID": u.bytes,
                    "RequestID": u.bytes,
                    "PlayerUniqueID": 0,
                },
                "Internal": False,
                "Version": 0x24,
                "UnLimited": False,
            },
        )
        return u

    def sendwscmd(self, cmd: str):
        u = uuid.uuid4()
        self.sendPacket(
            77,
            {
                "CommandLine": cmd,
                "CommandOrigin": {
                    "Origin": 5,
                    "UUID": u.bytes,
                    "RequestID": u.bytes,
                    "PlayerUniqueID": 0,
                },
                "Internal": False,
                "Version": 0x24,
                "UnLimited": False,
            },
        )
        return u

    def sendcmd_with_resp(
        self, cmd: str, timeout: float = 5
    ) -> Packet_CommandOutput | None:
        ud = self.sendcmd(cmd)
        getter, setter = utils.create_result_cb(dict)
        self.command_cbs[ud.bytes] = setter
        res = getter(timeout)
        del self.command_cbs[ud.bytes]
        if res is None:
            return None
        else:
            return Packet_CommandOutput(res)

    def sendwscmd_with_resp(
        self, cmd: str, timeout: float = 5
    ) -> Packet_CommandOutput | None:
        ud = self.sendwscmd(cmd)
        getter, setter = utils.create_result_cb(dict)
        self.command_cbs[ud.bytes] = setter
        res = getter(timeout)
        del self.command_cbs[ud.bytes]
        if res is None:
            return None
        else:
            return Packet_CommandOutput(res)

    def sendwocmd(self, cmd: str):
        self.sendPacket(
            140,
            {"CommandLine": cmd, "SuppressOutput": False},
        )

    @utils.thread_func("赞颂者消息处理", utils.ToolDeltaThread.SYSTEM)
    def handler(self, msg: Message) -> None:
        try:
            match msg.type:
                case MessageType.MSG_SET_BOT_BASIC_INFO:
                    self.bot_name = msg.content["bot_name"]
                    self.bot_uuid = msg.content["uuid"]
                    self.bot_unique_id = msg.content["bot_entity_unique_id"]
                    self.bot_runtime_id = msg.content["bot_runtime_id"]
                    self.bot_data_ready_event.set()
                case MessageType.MSG_UPDATE_UQ:
                    self.uqs = {k: PlayerUQ(**v) for k, v in msg.content.items()}
                    self.uq_data_ready_event.set()
                case MessageType.MSG_SERVER_PKT:
                    pkID = msg.content["ID"]
                    pk = msg.content["Content"]
                    if pkID == 79:
                        pkUUID = pk["CommandOrigin"]["UUID"]
                        if pkUUID in self.command_cbs.keys():
                            self.command_cbs[pkUUID](pk)
                        # else:
                        #    fmts.print_war(f"无效命令返回UUID: {pkUUID} ({self.command_cbs}) {id(self.command_cbs)}")
                    if self.packet_listener:
                        self.packet_listener(pkID, pk)
                case MessageType.MSG_CLIENT_PKT:
                    self.client_packet_handler(
                        msg.content["ID"], msg.content["Content"]
                    )
                case _:
                    fmts.print_war(f"未知数据传输类型: {msg.type}")
        except Exception as err:
            fmts.print_err(f"处理赞颂者通信出错: {err}")
            fmts.print_err(traceback.format_exc())

    def send(self, msg: Message):
        self.conn.send(msgpack.packb(msg.dumps())) # type: ignore



import struct
from dataclasses import dataclass
from io import BytesIO
from tooldelta.constants.packets import PacketIDS
from tooldelta.mc_bytes_packet.base_bytes_packet import BaseBytesPacket


@dataclass
class LevelChunk(BaseBytesPacket):
    Dimension: int = 0
    ChunkPosX: int = 0
    ChunkPosZ: int = 0
    HighestSubChunkIndex: int = 0
    CacheEnabled: bool = False

    def name(self) -> str:
        return "LevelChunk"

    def custom_packet_id(self) -> int:
        return 0

    def real_packet_id(self) -> int:
        return PacketIDS.IDLevelChunk

    def encode(self) -> bytes:
        raise NotImplementedError("Encode packet.LevelChunk is not support")

    def decode(self, bs: bytes):
        reader = BytesIO(bs)
        self.Dimension = reader.read(1)[0]
        self.ChunkPosX, self.ChunkPosZ = struct.unpack("<ii", reader.read(8))
        self.HighestSubChunkIndex = reader.read(1)[0]
        self.CacheEnabled = bool(reader.read(1)[0])



from typing import TypeVar
from collections.abc import Callable

VT = TypeVar("VT")


class Reader:
    def __init__(self, data: bytes):
        self.data = data
        self.offset = 0
        self.len = len(data)

    def read_byte(self):
        if self.offset >= self.len:
            raise ValueError(f"EOF: {self.offset}")
        value = self.data[self.offset]
        self.offset += 1
        return value

    def var_uint32(self):
        value = 0
        for i in range(5):
            byte = self.read_byte()
            value |= (byte & 0x7F) << (i * 7)
            if byte & 0x80 == 0:
                break
            elif i == 4:
                raise ValueError("VarUInt32 is too big")
        return value

    def uint8(self) -> int:
        return self.read_byte()

    def int8(self):
        return self.uint8() - 128

    def uint16(self):
        return self.uint8() | (self.uint8() << 8)

    def int16(self):
        return self.int8() | (self.int8() << 8)

    def bool(self) -> bool:
        value = self.uint8()
        return value == 1

    def string(self):
        strlen = self.var_uint32()
        value = self.data[self.offset : self.offset + strlen].decode()
        self.offset += strlen
        return value

    def bytes(self):
        length = self.var_uint32()
        value = self.data[self.offset : self.offset + length]
        self.offset += length
        return value

    def list(self, fnc: Callable[[], VT]) -> list[VT]:
        count = self.var_uint32()
        ls = []
        for _ in range(count):
            ls.append(fnc())
        return ls




from tooldelta.internal.launch_cli.neo_libs.blob_hash.client_function import (
    ClientFunction,
)
from tooldelta.internal.launch_cli.neo_libs.blob_hash.define import BaseBlobHashHolder
from tooldelta.internal.launch_cli.neo_libs.blob_hash.mirror_world_handler import (
    MirrorWorldHandler,
)
from tooldelta.internal.launch_cli.neo_libs.blob_hash.mirror_world_listener import (
    MirrorWorldListener,
)
from tooldelta.internal.launch_cli.neo_libs.neo_conn import ThreadOmega


class BlobHashHolder:
    """
    BlobHashHolder 是基于 BaseBlobHashHolder, ClientFunction, MirrorWorldHandler 和 MirrorWorldListener
    实现的，具有与 neomega-core 所拥有的 Blob hash holder 相同接口的 Blob hash cache 缓存数据集持有实现
    """

    _base_blob_hash_holder: BaseBlobHashHolder
    _client_function: ClientFunction
    _mirror_world_handler: MirrorWorldHandler
    _mirror_world_listener: MirrorWorldListener

    def __init__(self, omega: ThreadOmega) -> None:
        self._base_blob_hash_holder = BaseBlobHashHolder(omega, False, "")
        self._client_function = ClientFunction(self._base_blob_hash_holder)
        self._mirror_world_handler = MirrorWorldHandler(self._base_blob_hash_holder)
        self._mirror_world_listener = MirrorWorldListener(self._mirror_world_handler)

    def is_server(self) -> bool:
        """
        is_server 返回自身是否是 blob cache 缓存数据集的权威持有人(服务者)。
        这在 ToolDelta 中是不可能的，因此总是返回假
        """
        return False

    def is_disk_holder(self) -> bool:
        """is_disk_holder 返回当前是否是镜像存档资源的持有人"""
        return self._base_blob_hash_holder.is_disk_holder

    def wait_login_sequence_down(self):
        """
        wait_login_sequence_down 阻塞并等待登录序列完成。
        由于这是服务者所使用的函数，而这对于 ToolDelta 是不可能的，
        因此我们总是理解返回值而不进行任何操作
        """
        return

    def load_blob_cache(self, hash: int) -> bytes:
        """
        load_blob_cache 从底层缓存数据集检索 hash 所指示的数据负载。
        如果不存在，则返回空 bytes
        """
        return self._base_blob_hash_holder.omega.load_blob_cache(hash)

    def update_blob_cache(self, hash: int, payload: bytes) -> bool:
        """
        update_blob_cache 向底层缓存数据集写入 hash 所指示的二进制负载 payload。
        如果提供的 hash 并非 payload 的真实哈希，则写入将被拒绝，并返回假
        """
        return self._base_blob_hash_holder.omega.update_blob_cache(hash, payload)

    def as_server_side(self) -> None:
        """
        as_server_side 返回针对服务者实现的函数。
        非服务者调用 as_server_side 将得到空值。
        可以使用 b.IsServer() 确定当前结点是否是服务者。

        由于 ToolDelta 不可能成为服务者，所以该函数将立即返回值
        """
        return

    def as_mirror_world_side(self) -> MirrorWorldListener:
        """
        as_mirror_world_side 返回针对镜像存档资源持有者实现的 MirrorWorldListener，
        无论当前的使用者是否是镜像存档的持有人。

        通过 MirrorWorldListener.mirror_world_handler.set_handler 来设置处理函数，
        通过 MirrorWorldListener.register_listener 来启动监听
        """
        return self._mirror_world_listener

    def get_client_function(self) -> ClientFunction:
        """
        get_client_function 用于获取针对 Blob hash 客户端而实现的一系列函数，
        它是同时可被 服务者/客户端/镜像存档持有人 所共同使用的公共实现
        """
        return self._client_function



import struct

from dataclasses import dataclass, field
from io import BytesIO
from tooldelta.constants.packets import PacketIDS
from tooldelta.mc_bytes_packet.base_bytes_packet import BaseBytesPacket


@dataclass
class SubChunkRequest(BaseBytesPacket):
    Dimension: int = 0
    SubChunkPosX: int = 0
    SubChunkPosY: int = 0
    SubChunkPosZ: int = 0
    Offsets: list[tuple[int, int, int]] = field(default_factory=lambda: [])

    def name(self) -> str:
        return "SubChunkRequest"

    def custom_packet_id(self) -> int:
        return 2

    def real_packet_id(self) -> int:
        return PacketIDS.IDSubChunk

    def encode(self) -> bytes:
        writer = BytesIO()
        writer.write(self.Dimension.to_bytes(length=1, byteorder="little"))
        writer.write(
            struct.pack(
                "<ihiH",
                self.SubChunkPosX,
                self.SubChunkPosY,
                self.SubChunkPosZ,
                len(self.Offsets),
            )
        )
        for i in self.Offsets:
            writer.write(struct.pack("<bbb", i[0], i[1], i[2]))
        return writer.getvalue()

    def decode(self, bs: bytes):
        raise NotImplementedError("Decode packet.SubChunkRequest is not support")



"""
ToolDelta 基本框架

整个系统由三个部分组成
    Frame: 负责整个 ToolDelta 的基本框架运行
    GameCtrl: 负责对接游戏
        - Launchers: 负责将不同启动器的游戏接口统一成固定的接口，供插件在多平台游戏接口运行 (FastBuilder External, NeOmega, (BEWS, etc.))
    PluginGroup: 负责对插件进行统一管理
"""

import os
import signal
import sys
import traceback
import json

from . import constants, extend_functions, game_utils, utils
from .constants import SysStatus, TextType
from .internal.config_loader import ConfigLoader
from .internal.packet_handler import PacketHandler
from .internal.cmd_executor import ConsoleCmdManager
from .internal.maintainers.players import PlayerInfoMaintainer
from .internal.types import FrameExit
from .internal.launch_cli.neo_libs.blob_hash.blob_hash_holder import (
    BlobHashHolder,
)
from .internal.launch_cli import (
    FrameNeOmgAccessPoint,
    FrameNeOmegaLauncher,
    FB_LIKE_LAUNCHERS,
    LAUNCHERS,
)
from .game_texts import GameTextsHandle, GameTextsLoader
from .packets import Packet_CommandOutput
from .utils import cfg, fmts
from .version import get_tool_delta_version
from .plugin_load.plugins import PluginGroup


###### CONSTANT DEFINE
VERSION = get_tool_delta_version()


###### FRAME DEFINE
class ToolDelta:
    """ToolDelta 主框架"""

    class FrameBasic:
        """系统基本信息"""

        system_version = VERSION
        data_path = "插件数据文件/"

    def __init__(self) -> None:
        """初始化"""
        self.sys_data = self.FrameBasic()
        self.launchMode: int = 0
        self.on_plugin_err = staticmethod(
            lambda name, err: fmts.print_err(
                f"插件 <{name}> 出现问题: {err}\n§c{traceback.format_exc()}"
            )
        )

    def bootstrap(self):
        try:
            self.cfg_loader = ConfigLoader(self)
            self.welcome()
            self.init_dirs()
            self.packet_handler = PacketHandler(self)
            self.cmd_manager = ConsoleCmdManager(self)
            self.players_maintainer = PlayerInfoMaintainer(self)
            self.plugin_group = PluginGroup(self)
            self.game_ctrl = GameCtrl(self)
            self.add_console_cmd_trigger = self.cmd_manager.add_console_cmd_trigger
            self.launcher = self.cfg_loader.load_tooldelta_cfg_and_get_launcher()
            self.game_ctrl.hook_launcher(self.launcher)
            self.game_ctrl.hook_packet_handler(self.packet_handler)
            self.players_maintainer.hook_packet_handler(self.packet_handler)
            self.plugin_group.load_plugins()
            self.plugin_group.hook_packet_handler(self.packet_handler)
            self.launcher.set_packet_listener(self.packet_handler)
            utils.timer_event_boostrap()
            utils.tempjson.jsonfile_auto_save()
            game_utils.hook_packet_handler(self.packet_handler)
            extend_functions.load_functions(self)
            self.launcher.init()
            self.launcher.listen_launched(
                [
                    self.game_ctrl.system_inject,
                    self.cmd_manager.start_proc_thread,
                    extend_functions.activate_functions,
                ]
            )
            fmts.print_inf("正在唤醒游戏框架, 等待中...", end="\r")
            err = self.wait_closed()
            if not isinstance(err, SystemExit):
                fmts.print_err(f"启动器框架崩溃, 原因: {err}")
                return -1
            else:
                return 0
        except (KeyboardInterrupt, SystemExit, EOFError) as err:
            if str(err):
                fmts.print_inf(f"ToolDelta 已关闭，退出原因：{err}")
            else:
                fmts.print_inf("ToolDelta 已关闭")
        except Exception:
            fmts.print_err(f"ToolDelta 运行过程中出现问题：{traceback.format_exc()}")
        finally:
            self.system_exit("normal")

    def wait_closed(self):
        return self.launcher.launch()

    @staticmethod
    def welcome() -> None:
        """欢迎提示"""
        fmts.print_load("§dToolDelta Panel Embed By SuperScript")
        fmts.print_load("§dToolDelta Wiki: https://wiki.tooldelta.top/")
        fmts.print_load("§dToolDelta 交流群: 1030755163")
        fmts.print_load("§dToolDelta 项目地址: https://github.com/ToolDelta-Basic")
        fmts.print_load(f"§dToolDelta v {'.'.join([str(i) for i in VERSION])}")
        fmts.print_load("§dToolDelta Panel 已启动")

    def init_dirs(self):
        """初始化文件夹等"""
        os.makedirs(
            os.path.join("插件文件", constants.TOOLDELTA_CLASSIC_PLUGIN), exist_ok=True
        )
        os.makedirs("插件配置文件", exist_ok=True)
        os.makedirs(os.path.join("tooldelta", "bin"), exist_ok=True)
        os.makedirs(os.path.join("插件数据文件", "game_texts"), exist_ok=True)
        if sys.platform == "win32":
            self.win_create_batch_file()

    @staticmethod
    def win_create_batch_file():
        if not os.path.isfile("点我启动.bat"):
            argv = sys.argv.copy()
            if argv[0].endswith(".py"):
                argv.insert(0, "python")
            exec_cmd = " ".join(argv)
            with open("点我启动.bat", "w") as f:
                f.write(f"@echo off\n{exec_cmd}\npause")

    def system_exit(self, reason: str) -> None:
        """ToolDelta 系统退出"""
        # 启动器框架是否被载入
        has_launcher = hasattr(self, "launcher")
        if has_launcher:
            if self.launcher.status == SysStatus.RUNNING:
                self.launcher.update_status(SysStatus.NORMAL_EXIT)
        self.plugin_group.execute_frame_exit(
            FrameExit(self.launcher.status, reason), self.on_plugin_err
        )
        # 先将启动框架 (进程) 关闭了
        if has_launcher:
            if self.launcher.status == SysStatus.NORMAL_EXIT:
                # 运行中退出
                if isinstance(
                    self.launcher, FrameNeOmgAccessPoint | FrameNeOmegaLauncher
                ):
                    try:
                        self.game_ctrl.sendwocmd(
                            f'/kick "{self.game_ctrl.bot_name}" ToolDelta 退出中。'
                        )
                    except Exception:
                        pass
                    if not isinstance(self.launcher.neomg_proc, type(None)):
                        if os.name == "nt":
                            self.launcher.neomg_proc.send_signal(
                                signal.CTRL_BREAK_EVENT
                            )
                        else:
                            self.launcher.neomg_proc.send_signal(signal.SIGTERM)
            else:
                # 其他情况下退出 (例如启动失败)
                pass
            # 然后使得启动框架联络进程也退出
            if isinstance(self.launcher, FB_LIKE_LAUNCHERS):
                self.launcher.exit_event.set()
        # 最后做善后工作
        self.actions_before_exited()
        # 到这里就基本上是退出完成了

    def get_game_control(self) -> "GameCtrl":
        """获取 GameControl 对象

        Returns:
            GameCtrl: GameControl 对象
        """
        return self.game_ctrl

    def get_players(self) -> "PlayerInfoMaintainer":
        """获取 GameControl 对象

        Returns:
            GameCtrl: GameControl 对象
        """
        return self.players_maintainer

    @staticmethod
    def actions_before_exited() -> None:
        """安全退出"""
        utils.safe_close()
        fmts.print_inf("已保存数据与日志等信息。")

    def reload(self):
        """重载所有插件"""
        try:
            fmts.print_inf("重载: 正在保存数据缓存文件..")
            utils.safe_close()
            self.cmd_manager.reset_cmds()
            fmts.print_inf("重载: 正在重新载入插件..")
            self.plugin_group.reload()
            fmts.print_suc("重载插件: 全部插件重载成功！")
        except cfg.ConfigError as err:
            fmts.print_err(f"重载插件时发现插件配置文件有误: {err}")
        except SystemExit:
            fmts.print_err("重载插件遇到问题")
        except BaseException:
            fmts.print_err("重载插件遇到问题 (报错如下):")
            fmts.print_err(traceback.format_exc())
        finally:
            utils.timer_event_boostrap()


class GameCtrl:
    """游戏连接和交互部分"""

    def __init__(self, frame: "ToolDelta"):
        """初始化

        Args:
            frame (ToolDelta): 继承 Frame 的对象
        """
        try:
            self.game_texts_data = GameTextsLoader().game_texts_data
            self.game_data_handler = GameTextsHandle(self.game_texts_data)
        except Exception as err:
            fmts.print_war(f"游戏文本翻译器不可用: {err}")
            self.game_texts_data = None
            self.game_data_handler = None
        self.linked_frame = frame

    def hook_packet_handler(self, hdl: "PacketHandler"):
        hdl.add_dict_packet_listener(
            constants.PacketIDS.Text, self.handle_text_packet, -1
        )

    def hook_launcher(self, launcher: "LAUNCHERS"):
        self.launcher = launcher
        self.sendcmd = launcher.sendcmd
        self.sendwscmd = launcher.sendwscmd
        self.sendwocmd = launcher.sendwocmd
        self.sendPacket = launcher.sendPacket
        # TODO: pretty this code because it is not a good interface
        if hasattr(launcher, "blobHashHolder"):
            self.blobHashHolder = launcher.blobHashHolder

    def handle_text_packet(self, pkt: dict):
        """处理 文本 数据包

        Args:
            pkt (dict): 数据包内容
            plugin_grp (PluginGroup): 插件组对象
        """
        msg: str = pkt["Message"]
        match pkt["TextType"]:
            case TextType.TextTypeTranslation:
                if msg == "§e%multiplayer.player.joined":
                    playername = pkt["Parameters"][0]
                    fmts.print_inf(f"§e{playername} 加入了游戏")
                elif msg == "§e%multiplayer.player.left":
                    playername = pkt["Parameters"][0]
                    fmts.print_inf(f"§e{playername} 退出了游戏")
                else:
                    # TODO: another type of message
                    if self.game_data_handler is not None:
                        jon = self.game_data_handler.Handle_Text_Class1(pkt)
                        fmts.print_inf("§1" + " ".join(jon).strip('"'))
                    else:
                        fmts.print_inf(msg)
                    if msg.startswith("death."):
                        args = pkt["Parameters"]
                        utils.fill_list_index(args, ["", "", ""])
                        who_died, killer, weapon_name = args
                        if weapon_name:
                            fmts.print_inf(
                                f"§e{who_died} 被 {killer} 用 {weapon_name} §r杀死了"
                            )
                        elif killer:
                            fmts.print_inf(f"§e{who_died} 被 {killer} 击败了")
                        else:
                            fmts.print_inf(f"§e{who_died} 死亡了")
            case (
                TextType.TextTypeChat
                | TextType.TextTypeWhisper
                | TextType.TextTypeAnnouncement
            ):
                playername, message, ensurePlayer = (
                    utils.get_playername_and_msg_from_text_packet(
                        self.linked_frame, pkt
                    )
                )
                if playername is None or msg is None:
                    return False
                if ensurePlayer:
                    fmts.print_inf(f"<{playername}> {message}")
                else:
                    fmts.print_inf(f"<{playername}(伪造)> {message}")
            case TextType.TextTypeObjectWhisper:
                # /tellraw 消息
                msg = pkt["Message"]
                msg_text = json.loads(msg)["rawtext"]
                if len(msg_text) > 0 and msg_text[0].get("translate") == "***":
                    fmts.print_with_info("(该 tellraw 内容为敏感词)", "§f 消息 ")
                    return False
                msg_text = "".join([i["text"] for i in msg_text])
                fmts.print_with_info(msg_text, "§f 消息 ")
        return False

    def system_inject(self):
        self.linked_frame.players_maintainer.block_init()
        self.linked_frame.plugin_group.execute_init(self.linked_frame.on_plugin_err)
        self.inject_welcome()

    def inject_welcome(self) -> None:
        """初始化欢迎信息"""
        fmts.print_suc(
            "成功连接到游戏网络并初始化, 在线玩家: "
            + ", ".join(self.allplayers)
            + ", 机器人 ID: "
            + self.bot_name
        )
        self.sendcmd("/tag @s add robot")
        fmts.print_inf(
            "在控制台输入 §b插件市场§r 以§a一键获取§rToolDelta官方和第三方的插件"
        )
        fmts.print_suc("在控制台输入 §fhelp / ?§r§a 可查看控制台命令")
        if isinstance(self.launcher, FrameNeOmegaLauncher):
            fmts.print_suc("在控制台输入 o <neomega指令> 可执行NeOmega的控制台指令")

    # 向下兼容
    @property
    def players_uuid(self):
        return {
            v.name: k
            for k, v in self.linked_frame.players_maintainer.uuid_to_player.items()
        }

    # 向下兼容
    @property
    def allplayers(self):
        return [i.name for i in self.linked_frame.players_maintainer.getAllPlayers()]

    # 向下兼容
    @property
    def bot_name(self) -> str:
        if hasattr(self.launcher, "bot_name"):
            return self.launcher.get_bot_name()
        raise ValueError(
            f"此启动器 ({self.launcher.__class__.__name__}) 框架无法产生机器人名"
        )

    def sendcmd_with_resp(self, cmd: str, timeout: float = 30) -> Packet_CommandOutput:
        """
        发送普通指令并获取返回。

        Args:
            cmd (str): Minecraft 指令
            timeout (float, optional): 超时时间, 超时则引发 TimeoutError

        Returns:
            Packet_CommandOutput: 指令返回类
        """
        resp: Packet_CommandOutput = self.sendcmd(cmd, True, timeout)  # type: ignore
        return resp

    def sendwscmd_with_resp(
        self, cmd: str, timeout: float = 30
    ) -> Packet_CommandOutput:
        """
        发送 WebSocket 指令并获取返回。

        Args:
            cmd (str): MC WebSocket 指令
            timeout (float, optional): 超时时间, 超时则引发 TimeoutError

        Returns:
            Packet_CommandOutput: 指令返回类
        """
        resp: Packet_CommandOutput = self.sendwscmd(cmd, True, timeout)  # type: ignore
        return resp

    def say_to(self, target: str, text: str) -> None:
        """向玩家发送消息

        Args:
            target (str): 玩家名/目标选择器
            msg (str): 消息
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(f"tellraw {utils.to_player_selector(target)} {text_json}")

    def player_title(self, target: str, text: str) -> None:
        """向玩家展示标题文本

        Args:
            target (str): 玩家名/目标选择器
            text (str): 文本
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(f"titleraw {utils.to_player_selector(target)} title {text_json}")

    def player_subtitle(self, target: str, text: str) -> None:
        """向玩家展示副标题文本

        Args:
            target (str): 玩家名/目标选择器
            text (str): 文本
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(
            f"titleraw {utils.to_player_selector(target)} subtitle {text_json}"
        )

    def player_actionbar(self, target: str, text: str) -> None:
        """向玩家展示动作栏文本

        Args:
            target (str): 玩家名/目标选择器
            text (str): 文本
        """
        text_json = json.dumps({"rawtext": [{"text": text}]}, ensure_ascii=False)
        self.sendwocmd(
            f"titleraw {utils.to_player_selector(target)} actionbar {text_json}"
        )

    def get_game_data(self) -> dict:
        """获取游戏常见字符串数据

        Returns:
            dict: 游戏常见字符串数据
        """
        if self.game_texts_data is None:
            raise ValueError("游戏翻译器字符串数据不可用")
        return self.game_texts_data

    @property
    def players(self):
        """
        获取玩家信息存储 (PlayerInfoMaintainer)
        """
        return self.linked_frame.get_players()

    # TODO: pretty this code because it is not a good interface
    # 这个方法需要得到修改, 因为它不是一个通用接口
    def blob_hash_holder(self) -> BlobHashHolder:
        """
        blobHashHolder 返回 ToolDelta 的 Blob hash cache 缓存数据集的持有人。
        请确保当前的启动模式为 NeOmega 系启动器。

        Returns:
            BlobHashHolder: ToolDelta 的 Blob hash cache 缓存数据集的持有人
        """
        blobHashHolder: BlobHashHolder | None = getattr(self, "blobHashHolder", None)
        if blobHashHolder is None:
            raise ValueError("仅 NeOmega 系框架可使用 BlobHashHolder 功能")
        return self.blobHashHolder()



import time
import json
import os
import copy
from typing import Any, TypeVar
from collections.abc import Callable
from threading import Lock

from .timer_events import timer_event
from .safe_writer import safe_write

PathLike = str | os.PathLike[str]
VT = TypeVar("VT")

tempjson_rw_lock = Lock()
tempjson_paths: dict[PathLike, "_jsonfile_status"] = {}


class _jsonfile_status:
    def __init__(
        self,
        path: PathLike,
        need_file_exists: bool,
        default: Any = None,
        unload_delay: float | None = None,
    ):
        self.path = path
        self.is_changed = False
        self.load_time = time.time()
        self.unload_delay = unload_delay
        parent_dir = os.path.dirname(path)
        if parent_dir and not os.path.isdir(dp := os.path.dirname(path)):
            raise ValueError("文件夹: " + dp + " 路径不存在")
        if not need_file_exists and not os.path.isfile(path):
            with open(path, "w", encoding="utf-8") as f:
                json.dump(default, f, ensure_ascii=False)
            self.content = default
            self.is_changed = True
        else:
            with open(path, encoding="utf-8") as f:
                self.content = json.load(f)

    def flush_time(self):
        self.load_time = time.time()

    def should_unload(self):
        if self.unload_delay is None:
            return False
        return time.time() - self.load_time > self.unload_delay

    def read(self, deepcopy: bool = True):
        self.flush_time()
        if deepcopy:
            return copy.deepcopy(self.content)
        else:
            return self.content

    def write(self, content):
        self.flush_time()
        self.is_changed = content != self.content
        self.content = content

    def save(self):
        if self.is_changed:
            safe_write(self.path, self.content)




def load_from_path(
    path: PathLike,
    need_file_exists: bool = True,
    default: Any = None,
    unload_delay: int | None = None,
) -> _jsonfile_status:
    """
    将 json 文件从磁盘加载到缓存区，以便快速读写.
    在缓存文件已加载的情况下，再使用一次该方法不会有任何作用.

    Args:
        path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
        needFileExists (bool, optional): 默认为 True, 为 False 时，若文件路径不存在，就会自动创建一个文件
        default (Any, optional): 默认为 None, 为 None 时，若文件路径不存在，就会自动创建一个文件，且写入默认值 null
        unload_delay (int, optional): 多久没有再进行读写操作时, 将其从缓存卸载


    Raises:
        err: 文件不存在时
    """
    if j := tempjson_paths.get(path):
        return j
    j = tempjson_paths[path] = _jsonfile_status(
        path, need_file_exists=need_file_exists, default=default, unload_delay=unload_delay
    )
    return j


def unload_to_path(path: PathLike) -> bool:
    """
    将 json 文件从缓存区卸载 (保存内容到磁盘), 之后不能再在缓存区对这个文件进行读写.
    在缓存文件已卸载的情况下，再使用一次该方法不会有任何作用，但是可以通过其返回的值来知道存盘有没有成功.

    Args:
        path (str): 文件的虚拟路径

    Returns:
        bool: 存盘是否成功
    """
    if (jsonf := tempjson_paths.get(path)) is not None:
        jsonf.save()
        del tempjson_paths[path]
        return True
    return False


def read(path: PathLike, deepcopy: bool = True):
    """对缓存区的该虚拟路径的文件进行读操作，返回一个深拷贝的 JSON 对象

    Args:
        path (str): 文件的虚拟路径

    Raises:
        Exception: json 路径未初始化，不能进行读取和写入操作

    Returns:
        list[Any] | dict[Any, Any] | Any: 该虚拟路径的 JSON
    """
    if jsonf := tempjson_paths.get(path):
        return jsonf.read(deepcopy=deepcopy)
    raise ValueError(f"json 路径未初始化，不能进行读取和写入操作: {path}")


def get(path: PathLike) -> Any:
    """
    直接获取缓存区的该虚拟路径的 JSON, 不使用 copy
    WARNING: 如果你不知道有什么后果，请老老实实使用`read(...)`而不是`get(...)`!

    Args:
        path (str): 文件的虚拟路径
    """
    if jsonf := tempjson_paths.get(path):
        return jsonf.read(deepcopy=False)
    raise ValueError(f"json 路径未初始化, 不能进行读取和写入操作: {path}")


def write(path: PathLike, obj: Any) -> None:
    """
    对缓存区的该虚拟路径的文件进行写操作，这将会覆盖之前的内容

    Args:
        path (str): 文件的虚拟路径
        obj (Any): 任何合法的 JSON 类型 例如 dict/list/str/bool/int/float
    """
    if jsonf := tempjson_paths.get(path):
        jsonf.write(obj)
    else:
        raise ValueError(f"json 路径未初始化, 不能进行读取和写入操作：{path}")


def load_and_read(
    path: PathLike,
    need_file_exists: bool = True,
    timeout: int = 60,
    default: Callable[[], VT] | VT = {},
) -> VT:
    """读取 json 文件并将其从磁盘加载到缓存区，以便一段时间内能快速读写.

    Args:
        path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
        needFileExists (bool, optional): 默认为 True, 为 False 时，若文件路径不存在，就会自动创建一个文件，且写入默认值 null
        timeout (int, optional): 多久没有再进行读取操作时卸载缓存
        default (() -> Any | Any, optional):

    Returns:
        Any: 该虚拟路径的 JSON
    """
    if path not in tempjson_paths.keys():
        load_from_path(
            path,
            need_file_exists,
            default() if callable(default) else default,
            timeout,
        )
    return read(path)


def load_and_write(
    path: PathLike, obj: Any, need_file_exists: bool = True, timeout: int = 60
) -> None:
    """写入 json 文件并将其从磁盘加载到缓存区，以便一段时间内能快速读写.

    Args:
        path (str): 作为文件的磁盘内路径的同时也会作为在缓存区的虚拟路径
        obj (Any): 任何合法的 JSON 类型 例如 dict/list/str/bool/int/float
        needFileExists (bool, optional): 默认为 True, 为 False 时，若文件路径不存在，就会自动创建一个文件，且写入默认值 null
        timeout (int, optional): 多久没有再进行读取操作时卸载缓存
    """
    if path not in tempjson_paths.keys():
        load_from_path(path, need_file_exists, default=obj, unload_delay=timeout)
    write(path, obj)


def cancel_change(path: str):
    "取消缓存 json 所做的更改，非必要情况请勿调用，你不知道什么时候会自动保存所做更改"
    tempjson_paths[path].is_changed = False


def flush(path: PathLike | None = None):
    """
    刷新单个/全部JSON缓存区的缓存文件, 存入磁盘

    Args:
        path (str | None, optional): 文件虚拟路径, 默认全部存盘
    """
    if path:
        if tmpjson := tempjson_paths.get(path):
            tmpjson.save()
        else:
            raise ValueError(f"json 路径未初始化, 不能 flush: {path}")
    else:
        save_all()


def get_tmps() -> dict:
    "获取缓存区中所有临时文件"
    return tempjson_paths.copy()


@timer_event(30, "缓冲区 json 文件自动保存")
def jsonfile_auto_save():
    save_all()


def save_all():
    with tempjson_rw_lock:
        for k, v in tempjson_paths.copy().items():
            if v.should_unload():
                v.save()
                del tempjson_paths[k]
            else:
                v.save()


def reset():
    tempjson_paths.clear()



import os
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar, Any
from collections.abc import Callable

from ...mc_bytes_packet.pool import is_bytes_packet
from ...constants import TOOLDELTA_PLUGIN_DATA_DIR, PacketIDS
from ...utils import fmts
from ...internal.types import Player, Chat, InternalBroadcast, FrameExit
from . import event_cbs

if TYPE_CHECKING:
    from tooldelta import ToolDelta
    from tooldelta.internal.packet_handler import (
        DictPacketListener,
        BytesPacketListener,
    )
    from tooldelta.plugin_load.plugins import PluginGroup

    PLUGIN_TYPE = TypeVar("PLUGIN_TYPE", bound="Plugin")


class Plugin:
    "插件主类"

    name: str = ""
    "插件名"
    version = (0, 0, 1)
    "插件版本号"
    author: str = "?"
    "作者名"
    description = "..."
    "简介"

    # 如果将其作为 API 插件 ---
    _api_names: list[str] | None = None
    _api_ver: tuple[int, int, int] = (0, 0, 0)
    # ---

    # secret_vars ---
    __path_created__ = False
    _plugin_group: "PluginGroup | None" = None
    # ---

    def __init__(self, frame: "ToolDelta"):
        self.frame = frame
        self.game_ctrl = frame.get_game_control()

    @property
    def data_path(self) -> Path:
        "该插件的数据文件夹路径 (调用时直接创建数据文件夹)"
        path = os.path.join(TOOLDELTA_PLUGIN_DATA_DIR, self.name)
        if not self.__path_created__:
            os.makedirs(path, exist_ok=True)
            self.__path_created__ = True
        return Path(path)

    def make_data_path(self):
        os.makedirs(os.path.join(TOOLDELTA_PLUGIN_DATA_DIR, self.name), exist_ok=True)
        self.__path_created__ = True

    def print(self, msg: str):
        fmts.print_inf(f"{self.name}: {msg}")

    def format_data_path(self, *paths: str):
        return os.path.join(self.data_path, *paths)

    def ListenPreload(self, cb: Callable[[], Any]):
        """
        监听预加载事件
        预加载事件: 在读取插件后、和游戏建立连接前触发一次

        Args:
            cb (Callable[[], None]): 监听回调
        """
        event_cbs.on_preload_cbs.append((self, cb))

    def ListenActive(self, cb: Callable[[], Any]):
        """
        监听连接建立事件
        连接建立事件: 在框架和游戏完全建立连接时触发一次

        Args:
            cb (Callable[[], None]): 监听回调
        """
        event_cbs.on_active_cbs.append((self, cb))

    def ListenPlayerJoin(self, cb: Callable[[Player], Any]):
        """
        监听玩家加入事件
        玩家加入事件: 在有玩家加入游戏时触发一次

        Args:
            cb (Callable[[Player], None]): 监听回调, 传参: 玩家 (Player)
        """
        event_cbs.on_player_join_cbs.append((self, cb))

    def ListenPlayerLeave(self, cb: Callable[[Player], Any]):
        """
        监听玩家退出事件
        玩家退出事件: 在有玩家退出游戏时触发一次

        Args:
            cb (Callable[[Player], None]): 监听回调, 传参: 玩家 (Player)
        """
        event_cbs.on_player_leave_cbs.append((self, cb))

    def ListenChat(self, cb: Callable[[Chat], Any]):
        """
        监听玩家聊天事件
        玩家聊天事件: 在有玩家在聊天栏发言时触发一次

        Args:
            cb (Callable[[Player], None]): 监听回调, 传参: 聊天事件 (Chat)
        """
        event_cbs.on_chat_cbs.append((self, cb))

    def ListenFrameExit(self, cb: Callable[[FrameExit], Any]):
        """
        监听框架退出事件
        框架退出事件: 在框架退出/插件即将重载时触发一次

        Args:
            cb (Callable[[Player], None]): 监听回调, 传参: 聊天事件 (Chat)
        """
        event_cbs.on_frame_exit_cbs.append((self, cb))

    def ListenPacket(
        self, pkt_id: PacketIDS | list[PacketIDS], cb: "DictPacketListener"
    ):
        """
        监听字典数据包的事件

        Args:
            pkt_id (PacketIDS): 数据包ID
            cb ((dict) -> bool): 数据包监听回调, 返回 True 为拦截该数据包

        Raises:
            Exception: 尝试监听一个二进制数据包
        """
        if isinstance(pkt_id, int):
            pkt_ids = [pkt_id]
        else:
            pkt_ids = pkt_id

        for pkt_id in pkt_ids:
            if is_bytes_packet(pkt_id):
                raise Exception("你不能尝试使用 ListenPacket 监听二进制数据包")

        for pkt_id in pkt_ids:
            event_cbs.dict_packet_funcs.setdefault(pkt_id, [])
            event_cbs.dict_packet_funcs[pkt_id].append(cb)

    def ListenBytesPacket(
        self, pkt_id: PacketIDS | list[PacketIDS], cb: "BytesPacketListener"
    ):
        """
        监听二进制数据包的事件

        Args:
            pkt_id (PacketIDS): 数据包ID
            cb ((BytesPacketListener) -> bool): 数据包监听回调, 返回 True 为拦截该数据包

        Raises:
            Exception: 尝试监听一个非二进制数据包
        """
        if isinstance(pkt_id, int):
            pkt_ids = [pkt_id]
        else:
            pkt_ids = pkt_id

        for pkt_id in pkt_ids:
            if not is_bytes_packet(pkt_id):
                raise Exception("你不能尝试使用 ListenBytesPacket 监听普通的字典数据包")

        for pkt_id in pkt_ids:
            event_cbs.bytes_packet_funcs.setdefault(pkt_id, [])
            event_cbs.bytes_packet_funcs[pkt_id].append(cb)

    def ListenInternalBroadcast(
        self, broadcast_name: str, cb: Callable[[InternalBroadcast], Any]
    ):
        """
        监听广播事件
        广播事件: 在产生广播事件时触发一次

        Args:
            broadcast_name (str): 广播事件名
            cb (Callable[[InternalBroadcast], Any]): 监听回调, 传参: 广播事件 (InternalBroadcast), 返回: 回传给发送者的内容
        """
        event_cbs.broadcast_listener.setdefault(broadcast_name, [])
        event_cbs.broadcast_listener[broadcast_name].append(cb)

    def BroadcastEvent(self, evt: InternalBroadcast):
        """
        向全局广播一个特定事件, 可以传入附加信息参数
        Args:
            evt (InternalBroadcast): 事件
        Returns:
             list[Any]: 收集到的数据的列表 (如果接收到广播的方法返回了数据的话)
        """
        if self._plugin_group is None:
            raise RuntimeError("不能在非运行环境调用此方法")
        return self._plugin_group.brocast_event(evt)

    def GetPluginAPI(
        self,
        api_name: str,
        min_version: tuple[int, int, int] = (0, 0, 0),
        force=True,
    ):
        """获取插件 API

        Args:
            api_name (str): 插件 API 名
            min_version (tuple, optional): API 所需的最低版本 (若不填则默认不检查最低版本)
            force: 若为 False, 则在找不到插件 API 时不报错而是返回 None

        Raises:
            PluginAPIVersionError: 插件 API 版本错误
            PluginAPINotFoundError: 无法找到 API 插件

        Returns:
            Plugin: 插件 API
        """
        return self.frame.plugin_group.get_plugin_api(
            apiName=api_name, min_version=min_version, force=force
        )

    def get_typecheck_plugin_api(self, api_cls: type["PLUGIN_TYPE"]) -> "PLUGIN_TYPE":
        """
        对外源导入 (import) 的 API 插件类进行类型实例化。
        可以使得你所使用的 IDE 对导入的插件 API 类进行识别和高亮其所含方法。
        请在 TYPE_CHECKING 的代码块下使用。

        Args:
            api_cls (type[_PLUGIN_CLS_TYPE]): 导入的 API 插件类

        Raises:
            ValueError: API 插件类未被注册

        Returns:
            _PLUGIN_CLS_TYPE: API 插件实例

        使用方法如下:
        ```python
            p_api = get_plugin_api("...")
            if TYPE_CHECKING:
                from outer_api import api_cls_xx
                p_api = get_typecheck_plugin_api(api_cls_xx)
        ```
        """
        raise NotImplementedError("仅能在 TYPE_CHECKING 代码块调用此方法")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: proto/response.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'proto/response.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14proto/response.proto\x12\x16\x66\x61teark.proto.response\"\x98\x01\n\x0fGeneralResponse\x12>\n\x06status\x18\x01 \x01(\x0e\x32..fateark.proto.response.GeneralResponse.Status\x12\x0f\n\x07payload\x18\x02 \x01(\t\x12\x11\n\terror_msg\x18\x03 \x01(\t\"!\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\"\xa2\x01\n\x14GeneralInt32Response\x12\x43\n\x06status\x18\x01 \x01(\x0e\x32\x33.fateark.proto.response.GeneralInt32Response.Status\x12\x0f\n\x07payload\x18\x02 \x01(\x05\x12\x11\n\terror_msg\x18\x03 \x01(\t\"!\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\"\xa2\x01\n\x14GeneralInt64Response\x12\x43\n\x06status\x18\x01 \x01(\x0e\x32\x33.fateark.proto.response.GeneralInt64Response.Status\x12\x0f\n\x07payload\x18\x02 \x01(\x03\x12\x11\n\terror_msg\x18\x03 \x01(\t\"!\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\"\xa4\x01\n\x15GeneralUint64Response\x12\x44\n\x06status\x18\x01 \x01(\x0e\x32\x34.fateark.proto.response.GeneralUint64Response.Status\x12\x0f\n\x07payload\x18\x02 \x01(\x04\x12\x11\n\terror_msg\x18\x03 \x01(\t\"!\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\"\xa0\x01\n\x13GeneralBoolResponse\x12\x42\n\x06status\x18\x01 \x01(\x0e\x32\x32.fateark.proto.response.GeneralBoolResponse.Status\x12\x0f\n\x07payload\x18\x02 \x01(\x08\x12\x11\n\terror_msg\x18\x03 \x01(\t\"!\n\x06Status\x12\x0b\n\x07SUCCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\x42!Z\x1f./proto/gen/response;responsepbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.response_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\037./proto/gen/response;responsepb'
  _globals['_GENERALRESPONSE']._serialized_start=49
  _globals['_GENERALRESPONSE']._serialized_end=201
  _globals['_GENERALRESPONSE_STATUS']._serialized_start=168
  _globals['_GENERALRESPONSE_STATUS']._serialized_end=201
  _globals['_GENERALINT32RESPONSE']._serialized_start=204
  _globals['_GENERALINT32RESPONSE']._serialized_end=366
  _globals['_GENERALINT32RESPONSE_STATUS']._serialized_start=168
  _globals['_GENERALINT32RESPONSE_STATUS']._serialized_end=201
  _globals['_GENERALINT64RESPONSE']._serialized_start=369
  _globals['_GENERALINT64RESPONSE']._serialized_end=531
  _globals['_GENERALINT64RESPONSE_STATUS']._serialized_start=168
  _globals['_GENERALINT64RESPONSE_STATUS']._serialized_end=201
  _globals['_GENERALUINT64RESPONSE']._serialized_start=534
  _globals['_GENERALUINT64RESPONSE']._serialized_end=698
  _globals['_GENERALUINT64RESPONSE_STATUS']._serialized_start=168
  _globals['_GENERALUINT64RESPONSE_STATUS']._serialized_end=201
  _globals['_GENERALBOOLRESPONSE']._serialized_start=701
  _globals['_GENERALBOOLRESPONSE']._serialized_end=861
  _globals['_GENERALBOOLRESPONSE_STATUS']._serialized_start=168
  _globals['_GENERALBOOLRESPONSE_STATUS']._serialized_end=201
# @@protoc_insertion_point(module_scope)



import struct
import numpy

from dataclasses import dataclass, field
from io import BytesIO
from tooldelta.constants.packets import PacketIDS
from tooldelta.mc_bytes_packet.base_bytes_packet import BaseBytesPacket

SUB_CHUNK_RESULT_SUCCESS = 1
SUB_CHUNK_RESULT_CHUNK_NOT_FOUND = 2
SUB_CHUNK_RESULT_INVALID_DIMENSION = 3
SUB_CHUNK_RESULT_PLAYER_NOT_FOUND = 4
SUB_CHUNK_RESULT_INDEX_OUT_OF_BOUNDS = 5
SUB_CHUNK_RESULT_SUCCESS_ALL_AIR = 6


@dataclass
class SubChunkEntry:
    Result: int = 0
    SubChunkPosX: int = 0
    SubChunkPosY: int = 0
    SubChunkPosZ: int = 0
    NBTData: numpy.ndarray = field(
        default_factory=lambda: numpy.array([], dtype=numpy.uint8)
    )
    BlobHash: int = 0

    def decode(self, reader: BytesIO):
        self.Result = reader.read(1)[0]
        self.SubChunkPosX, self.SubChunkPosY, self.SubChunkPosZ = struct.unpack(
            "<ihi", reader.read(10)
        )
        self.NBTData = numpy.frombuffer(
            reader.read(struct.unpack("<I", reader.read(4))[0]), dtype=numpy.uint8
        )
        self.BlobHash = struct.unpack("<Q", reader.read(8))[0]


@dataclass
class SubChunk(BaseBytesPacket):
    Dimension: int = 0
    Entries: list[SubChunkEntry] = field(default_factory=lambda: [])
    CacheEnabled: bool = False

    def name(self) -> str:
        return "SubChunk"

    def custom_packet_id(self) -> int:
        return 0

    def real_packet_id(self) -> int:
        return PacketIDS.IDSubChunk

    def encode(self) -> bytes:
        raise NotImplementedError("Encode packet.SubChunk is not support")

    def decode(self, bs: bytes):
        reader = BytesIO(bs)
        self.Dimension = reader.read(1)[0]
        for _ in range(struct.unpack("<H", reader.read(2))[0]):
            s = SubChunkEntry()
            s.decode(reader)
            self.Entries.append(s)
        self.CacheEnabled = bool(reader.read(1)[0])



from typing import TYPE_CHECKING
from importlib import import_module
from .basic import ExtendFunction

if TYPE_CHECKING:
    from .. import ToolDelta

function_modules = [
    "gamerule_warnings",
]
registered_function_clss: list[type[ExtendFunction]] = []
registered_functions: list[ExtendFunction] = []


def regist_extend_function(func: type[ExtendFunction]):
    registered_function_clss.append(func)


def load_functions(frame: "ToolDelta"):
    for func_cls in registered_function_clss:
        func = func_cls(frame)
        registered_functions.append(func)


def activate_functions():
    for func in registered_functions:
        func.when_activate()


def import_functions():
    for function_module in function_modules:
        import_module(f".{function_module}", package=__package__)


import_functions()



"""配置文件模块"""

import os
import json
from typing import Any
from ..constants import TOOLDELTA_PLUGIN_CFG_DIR


NoneType = type(None)
VERSION = tuple[int, int, int]

PLUGINCFG_DEFAULT = {"配置版本": "0.0.1", "配置项": None}
PLUGINCFG_STANDARD_TYPE = {"配置版本": str, "配置项": [type(None), dict]}


def _cfg_isinstance_single(obj: Any, typ: type) -> bool:
    if not isinstance(typ, type):
        raise TypeError(f"_cfg_isinstance arg 1 must be a type, not {typ}")
    return {
        PInt: lambda: isinstance(obj, int) and obj > 0,
        NNInt: lambda: isinstance(obj, int) and obj >= 0,
        PFloat: lambda: isinstance(obj, float) and obj > 0,
        NNFloat: lambda: (isinstance(obj, float) or obj == 0) and obj >= 0,
        PNumber: lambda: isinstance(obj, int | float) and obj > 0,
        NNNumber: lambda: isinstance(obj, int | float) and obj >= 0,
    }.get(typ, lambda: isinstance(obj, typ))()


def _cfg_isinstance(obj: Any, typ: type | tuple[type]):
    """
    专用于 Cfg 的类型检测

    Args:
        obj (Any): 待检测对象
        typ (Any | tuple[Any]): 类型或类型元组

    Returns:
        bool: 是否为对应类型
    """
    if isinstance(typ, type):
        return _cfg_isinstance_single(obj, typ)
    if isinstance(typ, tuple):
        try:
            return any(_cfg_isinstance_single(obj, i) for i in typ)
        except TypeError as e:
            raise ValueError(f"_cfg_isinstance arg 2 can't be: {typ}") from e
    else:
        raise ValueError(f"_cfg_isinstance arg 2 can't be: {typ}")


def _get_cfg_type_name(typ: Any) -> str:
    """转换类型为中文字符串

    Args:
        typ (Any): 类型

    Returns:
        str: 中文字符串
    """
    if not isinstance(typ, type):
        typ = type(typ)
    return {
        PInt: "正整数",
        NNInt: "非负整数",
        PFloat: "正浮点小数",
        NNFloat: "非负浮点小数",
        str: "字符串",
        float: "浮点小数",
        int: "整数",
        dict: "json 对象",
        list: "列表",
        bool: "true/false",
        NoneType: "null",
    }.get(typ, typ.__name__)


class ConfigError(Exception):
    """配置文件错误"""

    def __init__(self, errStr: str, errPos: list | None = None):
        if errPos is None:
            errPos = []
        self.errPos = errPos
        self.args = (errStr,)


class JsonList:
    """配置文件的列表类型

    Args:
        patt: 判定规则
        len_limit (int): 限制列表的特定长度, 不限制则为-1
    """

    def __init__(self, patt: Any, len_limit=-1):
        self.patt = patt
        self.len_limit = len_limit


class AnyKeyValue:
    """配置文件的任意键名键值对类型"""

    def __init__(self, val_type: Any):
        self.type = val_type


class KeyGroup:
    """配置文件的键组，充当 dict_key 使用"""

    def __init__(self, *keys: str):
        self.keys = keys


class ConfigKeyError(ConfigError):
    """配置 json 的键错误"""


class ConfigValueError(ConfigError):
    """配置 json 的值错误"""


class VersionLowError(ConfigError):
    """配置 json 的版本过低的错误"""


class PInt(int):
    """配置文件的值限制：正整数"""


class NNInt(int):
    """配置文件的值限制：非负整数"""


class PFloat(float):
    """配置文件的值限制：正浮点小数"""


class NNFloat(float):
    """配置文件的值限制：非负浮点小数"""


class IntRange:
    """配置文件的值限制：整数域范围"""

    def __init__(self, min: int, max: int):
        self.min = min
        self.max = max


class FloatRange:
    """配置文件的值限制：浮点数域范围"""

    def __init__(self, min: float, max: float):
        self.min = min
        self.max = max


class PNumber:
    """配置文件的值限制：正数"""


class NNNumber:
    """配置文件的值限制：大于 0 的数"""


class FindNone:
    """找不到值"""


def get_cfg(path: str, standard_type: Any):
    """从 path 路径获取 json 文件文本信息，并按照 standard_type 给出的标准形式进行检测。"""
    path = path if path.endswith(".json") else f"{path}.json"
    with open(path, encoding="utf-8") as f:
        try:
            obj = json.load(f)
        except json.JSONDecodeError as exc:
            raise ConfigValueError(
                "JSON 配置文件格式不正确，请修正或直接删除", None
            ) from exc
    check_dict(standard_type, obj)
    return obj


def get_plugin_config_and_version(
    plugin_name: str,
    standard_type: Any,
    default: dict,
    default_vers: VERSION,
) -> tuple[dict[str, Any], VERSION]:
    """
    获取插件配置文件及版本

    Args:
        plugin_name (str): 插件名
        standard_type (dict): 标准类型
        default (dict): 默认配置
        default_vers (tuple[int, int, int]): 默认版本

    Returns:
            tuple[dict[str, Any], tuple[int, int, int]]: 配置文件内容及版本
    """
    # 详情见 插件编写指南.md
    assert isinstance(standard_type, dict)
    p = os.path.join(TOOLDELTA_PLUGIN_CFG_DIR, plugin_name)
    if not _jsonfile_exists(p) and default:
        defaultCfg = PLUGINCFG_DEFAULT.copy()
        defaultCfg["配置项"] = default
        defaultCfg["配置版本"] = ".".join([str(n) for n in default_vers])
        check_auto(standard_type, default)
        default_cfg(f"{p}.json", defaultCfg, force=True)
    cfg_stdtyp = PLUGINCFG_STANDARD_TYPE.copy()
    cfg_stdtyp["配置项"] = standard_type
    cfgGet = get_cfg(p, cfg_stdtyp)
    cfgVers = tuple(int(c) for c in cfgGet["配置版本"].split("."))
    VERSION_LENGTH = 3  # 版本长度
    if len(cfgVers) != VERSION_LENGTH:
        raise ValueError("配置文件出错：版本出错")
    return cfgGet["配置项"], cfgVers


def upgrade_plugin_config(
    plugin_name: str,
    configs: dict,
    version: VERSION,
):
    """
    获取插件配置文件及版本

    Args:
        plugin_name (str): 插件名
        configs (dict): 配置内容
        default_vers (tuple[int, int, int]): 版本
    """
    p = os.path.join(TOOLDELTA_PLUGIN_CFG_DIR, plugin_name)
    defaultCfg = PLUGINCFG_DEFAULT.copy()
    defaultCfg["配置项"] = configs
    defaultCfg["配置版本"] = ".".join([str(n) for n in version])
    write_default_cfg_file(f"{p}.json", defaultCfg, force=True)


def check_auto(
    standard: Any,
    val: Any,
    fromkey: str = "?",
):
    """
    检测任意类型的 json 类型是否合法

    Args:
        standard (type, dict, list): 标准模版
        val (Any): 待检测的值
        fromkey (Any, optional): 从哪个json键向下检索而来

    Raises:
        ValueError: 未知标准检测类型
        ConfigValueError: 值错误
    """
    if fromkey == FindNone:
        raise ValueError("不允许传入 FindNone")
    if isinstance(standard, type):
        if not _cfg_isinstance(val, standard):
            if isinstance(val, dict):
                raise ConfigValueError(
                    f'JSON 键"{fromkey}" 对应值的类型不正确：需要 {_get_cfg_type_name(standard)}, '
                    f"实际上为 json 对象：{json.dumps(val, ensure_ascii=False)}"
                )
            raise ConfigValueError(
                f'JSON 键"{fromkey}" 对应值的类型不正确：需要 {_get_cfg_type_name(standard)}, 实际上为 {_get_cfg_type_name(val)}'
            )
    elif isinstance(standard, JsonList):
        check_list(standard, val, fromkey)
    elif isinstance(standard, dict | AnyKeyValue):
        check_dict(standard, val, fromkey)
    elif isinstance(standard, IntRange):
        check_auto(int, val, fromkey)
        if not standard.min <= val <= standard.max:
            raise ConfigValueError(
                f'JSON 键"{fromkey}" 对应值的范围不正确：需要 {standard.min} ~ {standard.max}, 实际上为 {val}'
            )
    elif isinstance(standard, FloatRange):
        check_auto(float, val, fromkey)
        if not standard.min <= val <= standard.max:
            raise ConfigValueError(
                f'JSON 键"{fromkey}" 对应值的范围不正确：需要 {standard.min} ~ {standard.max}, 实际上为 {val}'
            )
    elif isinstance(standard, tuple | list):
        errs = []
        for single_type in standard:
            try:
                check_auto(single_type, val, fromkey)
                break
            except Exception as err:
                errs.append(err)
        else:
            reason = "\n".join(str(err) for err in errs)
            raise ConfigValueError(
                f'JSON 键 对应的键"{fromkey}" 类型不正确，以下为可能的原因：\n{reason}'
            )
    else:
        raise ValueError(f'JSON 键 "{fromkey}" 自动检测的标准类型传入异常：{standard}')


def auto_to_std(cfg):
    """
    从默认的配置文件内容的字典自动生成检测模版
    注意: 无法自动检测 AnyKeyValue, KeyGroup, PInt, PFloat 等

    Args:
        cfg: 默认的配置文件内容的字典

    Returns:
        标准检测模版样式, 用于 check_dict
    """
    if isinstance(cfg, dict):
        res = {}
        for k, v in cfg.items():
            if isinstance(v, dict | list):
                res[k] = auto_to_std(v)
            elif isinstance(v, str | int | float | bool):
                res[k] = type(v)
        return res
    if isinstance(cfg, list):
        setting_types = []
        for v in cfg:
            t = auto_to_std(v) if isinstance(v, dict | list) else type(v)
            if t not in setting_types:
                setting_types.append(t)
        if len(setting_types) == 1:
            return JsonList(setting_types[0])
        return JsonList(tuple(setting_types))
    raise ValueError("auto_to_std() 仅接受 dict 与 list 参数")


def check_dict(pattern: Any, jsondict: Any, from_key="?"):
    """
    按照给定的标准配置样式比对传入的字典, 键值对不上模版则引发相应异常
    请改为使用 check_auto

    参数:
        pattern: 标准模版 dict
        jsondict: 待检测的配置文件 dict
    """
    if not isinstance(jsondict, dict):
        raise ValueError(
            f'json 键"{from_key}" 需要 json 对象，而不是 {_get_cfg_type_name(jsondict)}'
        )
    if isinstance(pattern, AnyKeyValue):
        for key, val in jsondict.items():
            check_auto(pattern.type, val, key)
    else:
        for key, std_val in pattern.items():
            if isinstance(key, KeyGroup):
                for k, v in jsondict.items():
                    if k in key.keys:
                        check_auto(std_val, v, k)
            elif isinstance(key, str):
                val_get = jsondict.get(key, FindNone)
                if val_get == FindNone:
                    raise ConfigKeyError(f"不存在的 JSON 键：{key}")
                check_auto(std_val, val_get, key)
            else:
                raise ValueError(f"Invalid key type: {key.__class__.__name__}")


def check_list(pattern: JsonList, value: Any, fromkey: Any = "?") -> None:
    """
    检查列表是否合法
    请改为使用 check_auto

    Args:
        pattern (list): 标准模版
        value (Any): 待检测值
        fromkey (Any, optional): 从哪个字典键向下检索而来

    Raises:
        ValueError: 不是合法的标准列表检测样式
        ValueError: 标准检测列表的长度不能为 0
        ConfigValueError: 值错误
    """
    if not isinstance(pattern, JsonList):
        raise ValueError("不是合法的标准列表检测样式")
    if not isinstance(value, list):
        raise ConfigValueError(
            f'JSON 键 "{fromkey}" 需要列表 而不是 {_get_cfg_type_name(value)}'
        )
    if pattern.len_limit != -1 and len(value) != pattern.len_limit:
        raise ConfigValueError(
            f'JSON 键 "{fromkey}" 所对应的值列表有误：需要 {pattern.len_limit} 项，实际上为 {len(value)} 项'
        )
    for val in value:
        check_auto(pattern.patt, val, fromkey)


def write_default_cfg_file(path: str, default: dict, force: bool = False) -> None:
    """
    生成默认配置文件

    Args:
        path (str): 路径
        default (dict): 默认配置
        force (bool, optional): 是否即使是配置文件存在时, 也强制覆盖内容.
    """
    path = path if path.endswith(".json") else f"{path}.json"
    if force or not os.path.isfile(path):
        with open(path, "w", encoding="utf-8") as f:
            json.dump(default, f, indent=4, ensure_ascii=False)


default_cfg = write_default_cfg_file


def _jsonfile_exists(path: str) -> bool:
    return os.path.isfile(path if path.endswith(".json") else f"{path}.json")
